/*
 * Bundesliga is Copyright (c) 1997-2000 by Thomas Pundt
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose, without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice and this paragraph and the
 * following two paragraphs appear in all copies.
 *
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS,
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 * THE AUTHOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"
 * BASIS, AND THE AUTHOR HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#include <Common.h>
#include <System/SysAll.h>
#include <UI/UIAll.h>

#include "bl.h"
#include "blrcp.h"

/**
 * Database Handling Routinen
 */

static DmOpenRef BlDB;
static SelectedDBInfo dbInfo[20];
CharPtr Items[20] = { 0, };
static Err err = 0;
Short nrOfItems = 0;

static Short DBNameNr(CharPtr);
static ULong dbhash(CharPtr name);

/*
 * Create a new Database, install all records.
 */
DmOpenRef CreateDatabase(CharPtr name)
{
  UInt      index;
  DmOpenRef db;
  LocalID   dbID;
  VoidHand  p;

  if (DBNameNr(name)) {
    FrmAlert(ID_FrmDBExists);
    err = -1;
    return NULL;
  }

  if ( (err = DmCreateDatabase(0, name, 'tpBL', 'Data', false)) )
    return NULL;
  dbID = DmFindDatabase(0, name);
  db = DmOpenDatabase(0, dbID, dmModeReadWrite);

  index = 0;
  p = MemHandleLock(DmNewRecord(db, &index, sizeof(prefs)));
  DmWrite(p, 0, &prefs, sizeof(prefs));
  MemPtrUnlock(p);
  DmReleaseRecord(db, index, true);

  index = 1;
  DmNewRecord(db, &index, sizeof(results));
  DmReleaseRecord(db, index, true);

  index = 2;
  DmNewRecord(db, &index, sizeof(programme));
  DmReleaseRecord(db, index, true);

  index = 3;
  DmNewRecord(db, &index, sizeof(team));
  DmReleaseRecord(db, index, true);

  index = 4;
  DmNewRecord(db, &index, sizeof(Datum));
  DmReleaseRecord(db, index, true);

  return db;
}

/*
 * Read and store contents of Database db "locally".
 */
Err ReadDatabase(DmOpenRef db)
{
  VoidHand      RecHandle;
  VoidPtr	RecPointer;

  RecHandle = DmGetRecord(db, 0);
  RecPointer = MemHandleLock(RecHandle);
  MemMove(&prefs, RecPointer, sizeof(prefs));
  MemHandleUnlock(RecHandle);
  DmReleaseRecord(db, 0, true);
  
  RecHandle = DmGetRecord(db, 1);
  RecPointer = MemHandleLock(RecHandle);
  MemMove(results, RecPointer, sizeof(results));
  MemHandleUnlock(RecHandle);
  DmReleaseRecord(db, 1, true);
  
  RecHandle = DmGetRecord(db, 2);
  RecPointer = MemHandleLock(RecHandle);
  MemMove(programme, RecPointer, sizeof(programme));
  MemHandleUnlock(RecHandle);
  DmReleaseRecord(db, 2, true);
  
  RecHandle = DmGetRecord(db, 3);
  RecPointer = MemHandleLock(RecHandle);
  MemMove(team, RecPointer, sizeof(team));
  MemHandleUnlock(RecHandle);
  DmReleaseRecord(db, 3, true);
  
  RecHandle = DmGetRecord(db, 4);
  RecPointer = MemHandleLock(RecHandle);
  MemMove(Datum, RecPointer, sizeof(Datum));
  MemHandleUnlock(RecHandle);
  DmReleaseRecord(db, 4, true);

  if (DmNumRecords(db) == 6) {
    RecHandle = DmGetRecord(db, 5);
    RecPointer = MemHandleLock(RecHandle);
    MemMove(handicap, RecPointer, sizeof(handicap));
    MemHandleUnlock(RecHandle);
    DmReleaseRecord(db, 5, true);
  } else {
    MemSet(handicap, sizeof(handicap), 0);
  }

  return 0;
}

static Boolean isDB(UInt cardNo, LocalID dbID)
{
  int i;
  if (nrOfItems==0) 
    FindInstalledDatabases();
  for (i=0; i<nrOfItems; i++) {
    if ( (dbInfo[i].cardNo == cardNo) &&
         (dbInfo[i].dbID == dbID) )
      return true;
  }
  return false;
}
 

/*
 * Trys to open database dbID; if it doesn't exist,
 * a new database is created.
 */
Err OpenDatabase(UInt cardNo, LocalID dbID)
{
  BlDB = 0;

  if (dbID==0) {
    if (nrOfItems == 0) 
      FindInstalledDatabases();
    if (nrOfItems != 0) {
      cardNo = dbInfo[0].cardNo;
      dbID   = dbInfo[0].dbID;
    }
  }
  if (dbID && isDB(cardNo,dbID))
    BlDB = DmOpenDatabase(cardNo, dbID, dmModeReadWrite);

  if (!BlDB) {
    if (!(BlDB = CreateDatabase("Bundesliga 2000"))) {
      /* CreateDatabase sets err; err==-1 => our own error */
      return err;
    }

  } else {

    /* ReadDatabase also sets err; err==-1 => our own error */
    if (ReadDatabase(BlDB)) {
      return err;
    }
  }

  if (!prefs.Spieltage)
    prefs.Spieltage = 2*(prefs.NumTeams + (prefs.NumTeams%2) -1);

  return 0;
}

/*
 * Saves program status to open database and updates database
 * "nr" attributes.
 */
void SaveStatus(Short nr)
{
  VoidPtr p; 
  int i;
  
  prefs.matchDay = matchDay;
  
  p = MemHandleLock(DmResizeRecord(BlDB, 0, sizeof(prefs)));
  DmWrite(p, 0, &prefs, sizeof(prefs));
  MemPtrUnlock(p);
  DmReleaseRecord(BlDB, 0, true);

  p = MemHandleLock(DmResizeRecord(BlDB, 1, sizeof(results)));
  DmWrite(p, 0, results, sizeof(results));
  MemPtrUnlock(p);
  DmReleaseRecord(BlDB, 1, true);

  p = MemHandleLock(DmResizeRecord(BlDB, 2, sizeof(programme)));
  DmWrite(p, 0, programme, sizeof(programme));
  MemPtrUnlock(p);
  DmReleaseRecord(BlDB, 2, true);

  p = MemHandleLock(DmResizeRecord(BlDB, 3, sizeof(team)));
  DmWrite(p, 0, team, sizeof(team));
  MemPtrUnlock(p);
  DmReleaseRecord(BlDB, 3, true);

  p = MemHandleLock(DmResizeRecord(BlDB, 4, sizeof(Datum)));
  DmWrite(p, 0, Datum, sizeof(Datum));
  MemPtrUnlock(p);
  DmReleaseRecord(BlDB, 4, true);

  if (hasHandicap()) {
    UInt index = 5;
    if (DmNumRecords(BlDB) <6)
      p = MemHandleLock(DmNewRecord(BlDB, &index, sizeof(handicap)));
    else
      p = MemHandleLock(DmResizeRecord(BlDB, index, sizeof(handicap)));
    DmWrite(p, 0, handicap, sizeof(handicap));
    MemPtrUnlock(p);
    DmReleaseRecord(BlDB, index, true);
  }

  dbInfo[nr].dbAttr |= dmHdrAttrBackup;
  DmSetDatabaseInfo (dbInfo[nr].cardNo, dbInfo[nr].dbID, NULL, 
		     &dbInfo[nr].dbAttr, 
		     NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
  DmCloseDatabase(BlDB);

  if (appPrefs.options & dontSaveEM)
    appPrefs.options &= ~editMode;

  /* remember sort order the user chose for the leagues */
  for (i = 0; i<20; i++) {
    if (i<nrOfItems) {
      appPrefs.dbHash[i] = dbhash(Items[i]);
    } else {
      appPrefs.dbHash[i] = 0;
    }
  }

  appPrefs.Version = 1;
  PrefSetAppPreferences('tpBL', appPrefs.Version, 
			(void*)&appPrefs, sizeof(appPrefs));
}

/*
 * write "sane" values to freshly created database db
 */
void InitNewDatabase(DmOpenRef db, BlDatabasePreferencesType* newPrefs)
{
  Int i;
  VoidPtr p; 
  Word prg[520];
  char t[20][18];
  
  for (i=0; i<520; i++) {
    prg[i] = ((i%10)*2+1)*100 + ((i%10)*2+2);
  }
  for (i=0; i<20; i++) {
    StrIToA( t[i], i+1 );
  }

  p = MemHandleLock(DmResizeRecord(db, 0, sizeof(prefs)));
  DmWrite(p, 0, newPrefs, sizeof(prefs));
  MemPtrUnlock(p);
  DmReleaseRecord(db, 0, true);

  p = MemHandleLock(DmResizeRecord(db, 1, sizeof(results)));
  DmSet(p, 0, sizeof(results), 0xff);
  MemPtrUnlock(p);
  DmReleaseRecord(db, 1, true);

  p = MemHandleLock(DmResizeRecord(db, 2, sizeof(prg)));
  DmWrite(p, 0, prg, sizeof(prg));
  MemPtrUnlock(p);
  DmReleaseRecord(db, 2, true);

  p = MemHandleLock(DmResizeRecord(db, 3, sizeof(t)));
  DmWrite(p, 0, t, sizeof(t));
  MemPtrUnlock(p);
  DmReleaseRecord(db, 3, true);

  p = MemHandleLock(DmResizeRecord(db, 4, sizeof(Datum)));
  DmSet(p, 0, sizeof(Datum), 0);
  MemPtrUnlock(p);
  DmReleaseRecord(db, 4, true);

  nrOfItems = 0;
}

/*
 * looks for installed databases and stores result into global dbInfo[].
 */
Short FindInstalledDatabases()
{
  VoidHand  RecHandle;
  VoidPtr   RecPointer;
  DmOpenRef db;
  BlDatabasePreferencesType newPrefs;
  Boolean newSearch = true;
  DmSearchStateType stateInfo;
  Short i = 0;

  while (DmGetNextDatabaseByTypeCreator(newSearch, &stateInfo, 
	 'Data', 'tpBL', false, &dbInfo[i].cardNo, &dbInfo[i].dbID) == 0 
	 && i<20) {
    newSearch = false;

    db = DmOpenDatabase(dbInfo[i].cardNo, dbInfo[i].dbID, dmModeReadOnly);
    RecHandle = DmGetRecord(db, 0);
    RecPointer = MemHandleLock(RecHandle);
    MemMove(&newPrefs, RecPointer, sizeof(newPrefs));
    MemHandleUnlock(RecHandle);
    DmReleaseRecord(db, 0, true);
    DmCloseDatabase(db);
  
    if (newPrefs.Version<5) {
      continue;
    }
    i++;
  }
  nrOfItems = i;
  return nrOfItems;
}  

/*
 * returns number of database with name "name"
 */
static Short DBNameNr(CharPtr name)
{
  Short i;
  for (i=0; Items[i]; i++) {
    if (!StrCompare(name,Items[i]))
      return i+1;
  }
  return 0;
}

/* Sorting function for use with SysInsertionSort */
Int dbSortFunc(VoidPtr a, VoidPtr b, Long other)
{
  ULong keyA = dbhash(((SelectedDBInfo*)a)->dbName);
  ULong keyB = dbhash(((SelectedDBInfo*)b)->dbName);
  int i;

  for (i=0; i<20; i++) {
    if (keyA == appPrefs.dbHash[i])
      return -1;
    if (keyB == appPrefs.dbHash[i])
      return 1;
  }
  return 0;
}

void ChooseDB()
{
  Short i;
  ListPtr p;
    
  if (nrOfItems == 0)
    FindInstalledDatabases();

  for (i=0; i<nrOfItems; i++) {
    DmDatabaseInfo(dbInfo[i].cardNo, 
		   dbInfo[i].dbID, 
		   dbInfo[i].dbName, 
		   NULL, NULL, NULL, NULL, NULL, 
		   NULL, NULL, NULL, NULL, NULL);
  }

  /* perserve choosen sort order */
  SysInsertionSort(dbInfo,nrOfItems,sizeof(SelectedDBInfo),dbSortFunc,0);

  for (i=0; i<nrOfItems; i++) {
    Items[i] = i<nrOfItems?dbInfo[i].dbName:0;

    /* this must appear _after_ sorting the items. Having it here
     * doesn't hurt, because appPrefs.dbID is consistently updated
     */
    if ( (appPrefs.dbID == dbInfo[i].dbID) && 
	 (appPrefs.cardNo == dbInfo[i].cardNo) ) 
	 DBNr = i;
  }

  p = FrmGetListPtr(ID_FrmBl, ID_FrmPopup+1);
  LstSetListChoices(p, Items, nrOfItems);
  LstSetHeight(p, nrOfItems>10?10:nrOfItems);
}

void SelectedDB(Short i, SelectedDBInfo *info)
{
  MemMove(info, &dbInfo[i], sizeof(SelectedDBInfo));
  appPrefs.dbID = dbInfo[i].dbID;
  appPrefs.cardNo = dbInfo[i].cardNo;
}

void DeleteDatabase(Short i)
{
  Err err;
  Short j;
  err = DmCloseDatabase(BlDB);
  if ((err = DmDeleteDatabase(dbInfo[i].cardNo, dbInfo[i].dbID))) {
    FrmAlert(ID_FrmDBLoeschenError);
    return;
  }
  for (j=i; j<nrOfItems-1; j++) {
    MemMove(&dbInfo[j], &dbInfo[j+1], sizeof(dbInfo[j]));
  }
  nrOfItems--;
  ChooseDB();
}

Boolean RenameDatabase(Short i, CharPtr name)
{
  if (DmSetDatabaseInfo (dbInfo[i].cardNo, dbInfo[i].dbID, 
			 name, NULL, NULL, NULL, NULL, NULL, 
			 NULL, NULL, NULL, NULL, NULL) == 0) {
    StrCopy(dbInfo[i].dbName,name);
    return true;
  }
  return false;
}

/* 
 * Hash function that generates keys from database names. The keys rather
 * than the database names are stored in the application preferences.
 * A good idea? I don't know...
 */
static ULong dbhash(CharPtr name)
{
  ULong key = 0;
  Byte c;

  while ((c = *name++)) {
    key += c;
    key ^= (key << 24) | (key << 16) | (key << 8);
  }

  return key;
}

/*
 * exchange two dbInfo entries
 */
void swapDbInfo(int i, int j)
{
  SelectedDBInfo tmp = dbInfo[i];
  dbInfo[i] = dbInfo[j];
  dbInfo[j] = tmp;
}

